CPLEX 11.0.1: optimal solution; objective 88.578; 1 dual simplex iterations (0 in phase I)

Options
3
0
1
0
4
4
8
8
0
0
0.126
0
0
0
0
0
46.866666666666674
0
0
0
objno 0 0
suffix 0 8 8 0 0
sstatus
0 3
1 3
2 3
3 3
4 1
5 3
6 3
7 3
suffix 1 4 8 0 0
sstatus
0 1
1 1
2 3
3 1